// *******************************************************************************************************
// **                                                                                                   **
// **   AS7C34096B-10BIN_model.v - 4M Bits(512K X 8 BIT)  HIGH SPEED SRAM                               **
// **                                                                                                   **
// *******************************************************************************************************
// **                   This information is distributed under license from Alliance Memory              **
// **                              COPYRIGHT (c) 2017 Alliance Memory                                   **
// **                                      ALL RIGHTS RESERVED                                          **
// *******************************************************************************************************
// **                                                                                                   **
// **   Revision       : 0.1                                                                            **
// **   Modified Date  : 10/24/2017                                                                     **
// **   DataSheet file version :  AS7C34096B-10BIN-V1.5.pdf                       			**
// **   Revision History:                                                                               **
// **                                                                                                   **
// **   10/24/2017:  Initial design                                                                     **
// **																									                                                  **
// *******************************************************************************************************
`timescale 1ns/10ps 

module AS7C34096B-10BIN (A, DQ, CEn, OEn, WEn);
// *******************************************************************************************************
// **   Device define                                                                                   **
// *******************************************************************************************************
parameter DataWidth = 8;
parameter MemSize   = 524288;
parameter AddrWidth = 19;
// *******************************************************************************************************
// **   TIMING SPEC                                                                                     **
// ******************************************************************************************************* 

parameter delta   = 0.001;    // Delta for stimulus

parameter tRC   = 10;   // min.   Read Cycle Time
parameter tAA   = 10;   // max.   Address Access Time
parameter tACE  = 10;   // max.   Chip Enable Access Time
parameter tOE   = 4.5;  // max.   Output Enable Access Time
parameter tCLZ  = 2;    // min.   Chip Enable to Output in Low-Z
parameter tOLZ  = 0;    // min.   Output Enable to Output in Low-Z
parameter tCHZ  = 4;    // max.   Chip Disable to Output in High-Z
parameter tOHZ  = 4;    // max.   Output Disable to Output in High-Z
parameter tOH   = 2;    // min.   Output Hold from Address Change

parameter tWC   = 10;   // min.   Write Cycle Time
parameter tAW   = 8;    // min.   Address Valid to End of Write
parameter tCW   = 8;    // min.   Chip Enable to End of Write
parameter tAS   = 0;    // min.   Address Set-up time
parameter tWP   = 8;    // min.   Write Pulse Width
parameter tWR   = 0;    // min.   Write Recovery Time
parameter tDW   = 6;    // min.   Data to Write Time Overlap
parameter tDH   = 0;    // min.   Data Hold from End of Write Time
parameter tOW   = 2;    // min.   Output Active from End of Write
parameter tWHZ  = 4;    // max.   Write to Ouput in High-Z

//parameter for stimulus
parameter tRC_S  = tRC-delta;    // min.   Read Cycle Time
parameter tAA_S  = tAA-delta;    // max.   Address Access Time
parameter tACE_S = tACE-delta;   // max.   Chip Enable Access Time
parameter tOE_S  = tOE-delta;    // max.   Output Enable Access Time
parameter tCLZ_S = tCLZ-delta;   // min.   Chip Enable to Output in Low-Z
parameter tOLZ_S = tOLZ-delta;   // min.   Output Enable to Output in Low-Z
parameter tCHZ_S = tCHZ-delta;   // max.   Chip Disable to Output in High-Z
parameter tOHZ_S = tOHZ-delta;   // max.   Output Disable to Output in High-Z
parameter tOH_S  = tOH-delta;    // min.   Output Hold from Address Change

parameter tWC_S  = tWC-delta;    // min.   Write Cycle Time
parameter tAW_S  = tAW-delta;    // min.   Address Valid to End of Write
parameter tCW_S  = tCW-delta;    // min.   Chip Enable to End of Write
parameter tAS_S  = tAS-delta;    // min.   Address Set-up time
parameter tWP_S  = tWP-delta;    // min.   Write Pulse Width
parameter tWR_S  = tWR-delta;    // min.   Write Recovery Time
parameter tDW_S  = tDW-delta;    // min.   Data to Write Time Overlap
parameter tDH_S  = tDH-delta;    // min.   Data Hold from End of Write Time
parameter tOW_S  = tOW-delta;    // min.   Output Active from End of Write
parameter tWHZ_S = tWHZ-delta;   // max.   Write to Ouput in High-Z


// *******************************************************************************************************
// **   PIN                                                                                             **
// ******************************************************************************************************* 
input    CEn, OEn, WEn;
input   [(AddrWidth - 1) : 0] A;      //[18:0]A
inout   [(DataWidth - 1) : 0] DQ;     // DQ8
//---------------------------------------------------------------- 
reg     [(DataWidth - 1)   : 0] dout;    //[8:0] dout;
reg     [(DataWidth - 1) : 0] bank0 [0 : MemSize-1]; //[7:0]bank0 [0:524287]
wire    r_en;
wire    w_en;
wire    w_en2;
wire    ReadLowBusEn,ReadHighBusEn,DataOutEn,DataOutEn_S,UnknowDataOutEn,UnknowDataOutEn_S;
reg     rA1,rA2,rA3,rCE,rOE,rLB,rUB,rWE;
reg     flag_EndOfWrite;
wire    rA,rReadDataStrobe,wReadCycle_1,wReadCycle_delay;
reg     rReadCycleHold;
wire    wReadDataStrobe_WEn,wReadDataStrobe_LBn,wReadDataStrobe_UEn,DataOutValid;
// *******************************************************************************************************
// **   Initial                                                                                         **
// ******************************************************************************************************* 
initial
  begin
    rCE <= 1'b0;
    rOE <= 1'b0;
    rLB <= 1'b0;
    rUB <= 1'b0;
    rWE <= 1'b0;
    rA1 <= 1'b1;  
    rA2 <= 1'b1;  
    rA3 <= 1'b1;  
    flag_EndOfWrite <= 1'b1;
    rReadCycleHold  <= 1'b0;
  end
// *******************************************************************************************************
// **   Read                                                                                            **
// *******************************************************************************************************



always @(posedge DataOutEn or negedge wReadDataStrobe_WEn)
begin
  dout[(DataWidth - 1) :             0] = ReadLowBusEn  ? bank0[A]: 8'bz ; 

end  
  
always @(A)                             // Address Toggle inside Chip Enable
begin
  if (~CEn)   
  begin       
         rA1 <= 1'b1;         // Address toggle to High-Z data output timming 
    #tOH_S rA1 <= 1'b0;         
  end 
end
always @(A)                             // Address Toggle inside Chip Enable
begin
  if (~CEn)   
  begin       
         rA2 <= 1'b0; 
    #tAA_S rA2 <= 1'b1;         
  end 
end

wire wCE, wCE_LZ;
reg  wCE_tACE, wCE_tCHZ, wCE_tCLZ ;
always @(CEn) 
begin 
  wCE_tACE <= #tACE_S ~CEn;
  wCE_tCHZ <= #tCHZ_S ~CEn;
  wCE_tCLZ <= #tCLZ_S ~CEn;
end
assign wCE = wCE_tACE & wCE_tCHZ;
assign wCE_LZ = wCE_tCLZ & ~CEn ;

wire wOE, wOE_LZ;
reg  wOE_tOE, wOE_tOHZ, wOE_tOLZ ;
always @(OEn) 
begin 
  wOE_tOE  <= #tOE_S  ~OEn;
  wOE_tOHZ <= #tOHZ_S ~OEn;
  wOE_tOLZ <= #tOLZ_S ~OEn;
end
assign wOE = wOE_tOE & wOE_tOHZ;
assign wOE_LZ = wOE_tOLZ & ~(OEn | ~WEn);

reg  wWE_tOW, wWE_tWHZ;
always @(WEn)
begin
  wWE_tOW  <= #tOW_S WEn;
  wWE_tWHZ <= #tWHZ_S WEn;
end 

reg  wWE;
always @(WEn)
begin
  if (WEn) wWE <= #tOW_S WEn;
  else     wWE <= #tWHZ_S WEn;
end
 

always @(negedge WEn)
begin
  if (wReadCycle_1 & DataOutValid) 
  begin
    rReadCycleHold <= 1'b1; 
    #tWHZ_S rReadCycleHold <= 1'b0;
  end
end

assign #0.01 ReadLowBusEn =  ~rLB;
assign DataOutValid = ~rA & wCE & wOE & wWE;
assign DataOutEn = ~rA & wCE & wOE & (wWE | rReadCycleHold);
assign #0.1 wReadDataStrobe_WEn = ~CEn & ~OEn & ~WEn;
assign #0.1 wReadDataStrobe_En = ~CEn & ~OEn & ~WEn ;
assign UnknowDataOutEn = (wOE_LZ & wCE_LZ) & ~DataOutEn & (WEn & wWE_tOW);
assign #0.01 DataOutEn_S=DataOutEn;
assign #0.01 UnknowDataOutEn_S=UnknowDataOutEn;
assign DQ = DataOutEn_S ? dout : (UnknowDataOutEn_S ? 8'bx : 8'bz);
assign rA = ~(rA1 | rA2);

// *******************************************************************************************************
// **   Write                                                                                           **
// *******************************************************************************************************

reg    AddEnable1, AddEnable2 ,rLBC , rUBC;
wire   AddEnable;
reg    [7:0] DQ_pathdelay;
assign AddEnable = (AddEnable1 ^ AddEnable2);
assign w_en  =  (~WEn) & (~CEn);
assign w_en2 =  (~WEn) & (~CEn);	
always @(DQ) DQ_pathdelay  <= #0.1 DQ;

initial
begin
   AddEnable1 = 1'b0;
   AddEnable2 = 1'b0;
end


always @ (negedge w_en)
begin
	    bank0[A] = DQ_pathdelay [(DataWidth - 1) : 0];
end  

always @ (negedge w_en2)
begin
	    bank0[A] = DQ_pathdelay [(DataWidth - 1) : 0];	
end 

always @(A)
begin
  if (~CEn  & ~WEn )
  begin
    AddEnable1 <= ~AddEnable1;     
    AddEnable2 <= #tAW_S ~AddEnable2;
  end 
  else
  begin
    AddEnable1 <= 1'b0;  
    AddEnable2 <= 1'b0; 
  end 
end


// *******************************************************************************************************
// **   TIMING CHECKS                                                                                   **
// *******************************************************************************************************
//------------- tDW         : min.     6 Data to Write Time Overlap      : Valid data to End of write (WEn or CEn) min. time 6ns
  real tm_DQ_toggle1,tm_DQ_toggle2,tm_DQ_setup, tm_DQ_hold, tm_UB_rising, tm_LB_rising, tm_WE_rising, tm_CEn_rising, tm_CE2_falling, temp1,temp2,temp3;
  real tm_A_toggle, tm_A_last,flag_A_last;
  reg    dbg_;
  real   tm_CEn_falling,tm_CE2_rising;
  
  initial
  begin
    tm_A_toggle <=0;
    tm_A_last <=0;
    tm_DQ_toggle1 <= 0;
    tm_DQ_toggle2 <= 0;
    flag_A_last <= 0;
    dbg_ <= 1'b0;   
  end

  always @(negedge CEn) tm_CEn_falling <= $realtime;
    always @(negedge WEn) flag_EndOfWrite <= 1'b0;  
 
  always @(DQ) 
  begin
    tm_DQ_toggle1 <= $realtime;
    tm_DQ_toggle2 <= tm_DQ_toggle1;
  end
  always @(posedge WEn) 
  begin
    if (($realtime-tm_DQ_toggle1) > 0.1)
    begin
      if ($realtime - tm_DQ_toggle1 < tDW)
        $display("Error: tDW WEn rising DQ setup time violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10, $realtime - tm_DQ_toggle1,tDW);
    end
    else
    begin 
      if ($realtime - tm_DQ_toggle2 < tDW)
        $display("Error: tDW WEn rising DQ setup time violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10, $realtime - tm_DQ_toggle2,tDW);   
    end
  end
  always @(posedge CEn)
  begin
    if (($realtime-tm_DQ_toggle1) > 0.1)
    begin
      if ((~WEn) && ($realtime - tm_DQ_toggle1 < tDW))
        $display("Error: tDW CEn rising DQ setup time violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10, $realtime - tm_DQ_toggle1,tDW);
    end
    else
    begin 
      if ((~WEn) && ($realtime - tm_DQ_toggle2 < tDW))
        $display("Error: tDW CEn rising DQ setup time violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10, $realtime - tm_DQ_toggle2,tDW);     
    end
  end
//------------- tAS         : min.     0   Address Set-up time      : Address min. setup time 0ns, before CEn or WEn negedge
    parameter BusSkew =1.85;	
  wire  [(AddrWidth - 1) : 0]   A_glitch,A_delay ;
  reg                           rA_glitch;
  reg                           rCEnFalling_mask,rCE2Rising_mask,rWEnFalling_mask,rUBnFalling_mask,rLBnFalling_mask;
 

  initial
  begin
    rA_glitch =1'b0;
  end
  assign #0.1 A_delay   = A;                       //----------delay A 0.1ns to generate A transition pulse glitch
  assign      A_glitch  = A^A_delay;
  always @(A_glitch) 
  begin   
    rA_glitch <= 1'b1;
    #0.1 rA_glitch <= 1'b0;
  end
 
  always @(negedge WEn)
  begin
    if (~CEn)	
    begin
      #0.2 rWEnFalling_mask <= 1'b1;                      //--------------- 0.2ns filter out A trans align with CEn falling event
      #BusSkew rWEnFalling_mask <= 1'b0;
    end
  end

  always @(negedge CEn)
  begin
    #0.2 rCEnFalling_mask <= 1'b1;                          //--------------- 0.2ns filter out A trans align with CEn falling event
    #BusSkew rCEnFalling_mask <= 1'b0; 
  end 

  
  always @(rA_glitch)
  begin
    if (~CEn & ~WEn & rWEnFalling_mask & rA_glitch) $display("Error: tAS WEn falling to Address transition min. time violation at time %t ps", $realtime*10);	  
    if (~CEn & ~WEn & rCEnFalling_mask & rA_glitch) $display("Error: tAS CEn falling to Address transition min. time violation at time %t ps", $realtime*10);	
  end 
//---------------tWR:min.  0: Write Recovery Time before CEn or WEn posedge
  real tm_tran_tWR;
  always @(posedge rA_glitch)
  begin
    tm_tran_tWR <= $realtime;
  end 
  always @(posedge WEn)
  begin
	if ((~CEn) && ((BusSkew > $realtime-tm_tran_tWR)&&($realtime-tm_tran_tWR > 0))) 	
      $display("Error: tWR violation at WEn rising time %t ps:  Measure=%f ns Spec=%f ns", $realtime*10,tm_tran_tWR-$realtime,tWR);
  end

  always @(posedge CEn)
  begin
    if ((BusSkew > $realtime-tm_tran_tWR)&&($realtime-tm_tran_tWR > 0))
      $display("Error: tWR violation at CEn rising time %t ps:  Measure=%f ns Spec=%f ns", $realtime*10,tm_tran_tWR-$realtime,tWR);
  end 

//---------------tDH     min.    0   Data Hold time from End of Write(CEn or WEn )
  real  tm_tran_tDH;
  wire  [(DataWidth - 1) : 0]   DQ_glitch,DQ_delay ;
 
  assign #0.1 DQ_delay = DQ;                  //----------delay A 0.1ns to generate A transition pulse glitch
  assign      DQ_glitch = DQ^DQ_delay;  
  always @(posedge DQ_glitch)
  begin
    tm_tran_tDH  <= $realtime;
  end 
  always @(posedge WEn)
  begin
    flag_EndOfWrite <= 1'b1;
    if ((flag_EndOfWrite == 1'b0)&&(BusSkew > $realtime-tm_tran_tDH)&&($realtime-tm_tran_tDH > 0)) $display("Error: tDH WEn rising, tDH violation at time %t ps", $realtime*10);
  end 
  always @(posedge CEn)
  begin
    flag_EndOfWrite <= 1'b1;
    if ((flag_EndOfWrite == 1'b0)&&(~WEn) && (BusSkew > $realtime-tm_tran_tDH)&&($realtime-tm_tran_tDH > 0)) $display("Error: tDH CEn rising, tDH violation at time %t ps", $realtime*10);
  end 
 
//--------------- tRC and tWC : min. tRC=10 Read Cycle Time;  min. tWC=8 Write Cycle Time(Address transition min. width)
  real tm_A_tran_tWC_0, tm_A_tran_LastBit;
  reg  flag_tWCtRC,flag_tWCtRC1;
  initial
  begin
    tm_A_tran_tWC_0 <= 0;
    flag_tWCtRC <= 1'b0;
    flag_tWCtRC1 <= 1'b0;
  end

  always @(posedge CEn) flag_tWCtRC <= 1'b1;  
  always @(negedge CEn) flag_tWCtRC1 <= 1'b1;  
  always @(A)
  begin
    if (CEn) tm_A_tran_tWC_0 <= $realtime;	
    if (~CEn) flag_tWCtRC1 <= 1'b0;	
    if ((CEn) & flag_tWCtRC & flag_tWCtRC1)	
    begin
      flag_tWCtRC <= 1'b0;
      flag_tWCtRC1 <= 1'b0;
      if (((tWC-0.001) > $realtime-tm_A_tran_tWC_0) && ($realtime-tm_A_tran_tWC_0 > BusSkew))
      begin
        $display("Error: tWC/tRC A Cycle min. width violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10,$realtime-tm_A_tran_tWC_0,tRC);
      end
    end
    tm_A_tran_LastBit <= $realtime;
  end
//--------------- tRC/tWC   : min. tRC=10 Read Cycle Time; (Address transition min. width)
  real tm_A_tran_tRC;
  initial
  begin
    tm_A_tran_tRC <= 0;
  end
 
  always @(A)
  begin
    tm_A_tran_tRC <= $realtime;
    begin
      if (((tRC-0.001) > $realtime-tm_A_tran_tRC) && ($realtime-tm_A_tran_tRC > BusSkew))
      begin
        $display("Error: tWC/tRC A Cycle min. width violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10,$realtime-tm_A_tran_tRC,tRC);
        dbg_ <= ~dbg_;
      end
    end
  end
  
//------------- tAW         : min.     8 Address Valid to End of Write
  wire   wWriteCycle_delay;     //-- 20170529

  always @(posedge WEn)
  begin
    if (tAW > $realtime-tm_A_tran_LastBit) 
      $display("Error: tAW, Address transition to WEn rising violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10,$realtime-tm_A_tran_LastBit,tAW);
  end 
  always @(posedge CEn)
  begin
    if  ((wWriteCycle_delay) && (tAW > $realtime-tm_A_tran_LastBit))
      $display("Error: tAW, Address transition to CEn rising violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10,$realtime-tm_A_tran_LastBit,tAW); 
  end 
//------------- tCW         : min.     8 Chip Enable to End of Write 

  always @(posedge WEn)
  begin
    if (tCW > $realtime-tm_CEn_falling) $display("Error: tCW, CEn falling to WEn rising violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10,$realtime-tm_CEn_falling,tCW);
  end 
 
  always @(posedge CEn)
  begin
    if (tCW > $realtime-tm_CEn_falling) $display("Error: tCW, CEn falling to CEn rising violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10,$realtime-tm_CEn_falling,tCW);
  end   

//---------------- tWP   min. 8  Write Pulse Width 
  real tm_WEn_falling;
  always @(negedge WEn) tm_WEn_falling <= $realtime;
  always @(posedge WEn)
  begin
    if ((tWP-0.001) > $realtime-tm_WEn_falling) $display("Error: tWP min. WEn width violation at time %t ps: Measure=%f ns Spec=%f ns", $realtime*10,$realtime-tm_WEn_falling,tWP);
  end
//------------- tWHZ        : max.     4 Write to Ouput in High-Z 
    reg  rWEn_tWHZ,rUBn_tWHZ,rLBn_tWHZ,rCEn_tWHZ,rCE2_tWHZ;
  reg  rReadWriteCycle_tWHZ;
  reg  rReadCycle,rWriteCycle;
  initial 
  begin
    rWEn_tWHZ <= 1'b0;
    rUBn_tWHZ <= 1'b0;
    rLBn_tWHZ <= 1'b0;
    rCEn_tWHZ <= 1'b0;
    rCE2_tWHZ <= 1'b0;
    rReadWriteCycle_tWHZ <= 1'b0;
    rReadCycle <= 1'b0;
    rWriteCycle <= 1'b0;
  end
  always @(WEn) rWEn_tWHZ <= #tWHZ WEn; 

  always @(negedge WEn) if (OEn==1'b0) rReadWriteCycle_tWHZ <= 1'b1;
  else                                 rReadWriteCycle_tWHZ <= 1'b0;
  assign #(tWHZ+0.1) wReadCycle_delay = rReadCycle;     //-- 20170525
  assign #0.1        wReadCycle_1 = rReadCycle;
  assign #0.1        wWriteCycle_delay = rWriteCycle;
  
  always @(CEn or WEn or OEn)
  begin
    if (~CEn & WEn & ~OEn) rReadCycle <= 1'b1;
    else                   rReadCycle <= 1'b0;
    if (~CEn & ~WEn) rWriteCycle <= 1'b1;
    else             rWriteCycle <= 1'b0;
  end
  
  always @(negedge rWEn_tWHZ) if ((wReadCycle_delay == 1'b1) && (DQ[7:0] !== 8'hz)) $display("Error: tWHZ min. WEn falling to DQ High-Z min. time violation at time %t ps", $realtime*10);     //-- 20170525

endmodule
 
